const EventEmitter = require('events');
const dirTraveler = require('dir-traveler');
const Joi = require('joi');
const uuid = require('uuid');

const checkValidation = (data, schema) => {
  const { value, error } = Joi.validate(data, schema);
  if (error) throw error;
  return value;
};

class Poolishark extends EventEmitter {

  constructor(poolish, ctrlPath) {
    super();
    this._poolish = poolish;
    if (poolish.constructor.name === 'Server') {
      if (!ctrlPath) throw new Error('Poolishark server need the 2nd argument to be controllers path.');
      this._controllers = {};
      this._validControllers = [];
      this._clients = [];
      this._enhanceServer(ctrlPath);
      this._type = 'Server';
    } else if (poolish.constructor.name === 'Client') {
      this._enhanceClient();
      this._type = 'Client';
    }
  }

  get type() {
    return this._type;
  }

  get controllers() {
    return this._controllers;
  }

  get validControllers() {
    return this._validControllers;
  }

  get clients() {
    return this._clients;
  }

  async exec({ c, params = {} }) {
    const ts = Date.now();
    const controller = this._controllers[c];
    try {
      if (!controller) throw new Error(`Controller [${c}] has not registered.`);
      const result = await controller(params);
      this.emit('exec_succeed', c, params, Date.now() - ts);
      return result;
    } catch (error) {
      this.emit('exec_failed', c, params, error);
      throw error;
    }
  }

  _enhanceServer(ctrlPath) {
    const server = this._poolish;
    server.on('data', (payload, client) => {
      try {
        const { data, qid } = JSON.parse(payload);
        (async _ => {
          const result = await this.exec(data);
          if (qid) {
            client.send(JSON.stringify({ data: result, qid, succeed: true }));
          }
        })().catch(error => {
          if (qid) {
            client.send(JSON.stringify({
              error: {
                message: error.message,
                name: error.name,
                stack: error.stack,
              },
              qid,
              succeed: false,
            }));
          }
          this.emit('error', error, 'CONTROLLER');
        });
      } catch (error) {
        this.emit('error', error, 'SERVER_PARSE_PAYLOAD');
      }
    });
    server.on('listening', port => {
      this.emit('listening', port);
      this._initControllers(ctrlPath);
    });
    server.on('error', (error, tag) => this.emit('error', error, tag));
    server.on('connected', client => {
      this.emit('connected', client);
      this._clients.push(client);
    });
    server.on('disconnected', client => {
      this.emit('disconnected', client);
      this._clients.splice(this._clients.indexOf(client), 1);
    });
  }

  _initControllers(ctrlPath) {
    const files = dirTraveler(ctrlPath);
    this._controllers = Object.keys(files).reduce((memo, file) => {
      const controller = file.slice(0, file.lastIndexOf('.'));
      if (controller.slice(controller.lastIndexOf('/') + 1)[0] === '_') return memo;
      const { validate = Joi.any(), output = Joi.any(), handler = _ => null } = require(files[file]);
      if (!memo[controller]) {
        this._validControllers.push(controller);
      }
      memo[controller] = async params => {
        const result = await handler.bind(this)(checkValidation(params, validate));
        if (typeof result !== 'undefined') {
          return checkValidation(JSON.parse(JSON.stringify(result)), output);
        }
      };
      this.emit('reg', controller);
      return memo;
    }, {});
  }

  _enhanceClient() {
    const client = this._poolish;
    client._contexts = {};
    client.on('data', payload => {
      try {
        const { qid, data, succeed, error } = JSON.parse(payload);
        const context = client._contexts[qid];
        if (context) {
          clearTimeout(context.timer);
          Reflect.deleteProperty(client._contexts, qid);
          if (succeed) {
            context.resolve(data);
          } else {
            const err = new Error(error.message);
            err.name = error.name;
            err.stack = error.stack;
            context.reject(err);
          }
        }
      } catch (error) {
        this.emit('error', error, 'CLIENT_PARSE_PAYLOAD');
      }
    });
    client.on('connected', client => this.emit('connected', client));
    client.on('disconnected', client => this.emit('disconnected', client));
    client.on('error', (error, tag) => this.emit('error', error, tag));
    this.send = (data, { priority } = {}) => {
      client.send(JSON.stringify({ data }), { priority });
    };
    this.sendSync = async (data, { priority, timeout = 10000, detail = false } = {}) => {
      return new Promise((resolve, reject) => {
        const qid = uuid.v4();
        client._contexts[qid] = {
          resolve,
          reject,
          timer: setTimeout(_ => {
            Reflect.deleteProperty(client._contexts, qid);
            let log = data.c;
            if (detail) {
              log = JSON.stringify(data);
            }
            reject(new Error(`Timeout ${timeout}ms, ${log}`));
          }, timeout),
        };
        client.send(JSON.stringify({ data, sync: true, qid }), { priority });
      });
    };
  }
}

Poolishark.Joi = Joi;
module.exports = Poolishark;
